<?php
/**
 * The library immonex WP Free Plugin Core provides shared basic functionality
 * for free immonex WordPress plugins.
 * Copyright (C) 2014, 2022  inveris OHG / immonex
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *
 * This file contains the base class for deriving the main classes of
 * immonex plugins.
 *
 * @package immonex\WordPressFreePluginCore
 */

namespace immonex\WordPressFreePluginCore\DEV_8;

/**
 * Base class for free immonex WordPress plugins.
 *
 * @version 1.6.0
 */
abstract class Base {

	const BASE_VERSION = '1.6.0';

	/**
	 * Plugin options array
	 *
	 * @var mixed[]
	 */
	protected $plugin_options = array();

	/**
	 * Plugin options default values
	 *
	 * @var mixed[]
	 */
	protected $default_plugin_options = array();

	/**
	 * Minimun WP capability to access the plugin options page
	 *
	 * @var string
	 */
	protected $plugin_options_access_capability = 'manage_options';

	/**
	 * Does the plugin has its own options page?
	 *
	 * @var bool
	 */
	protected $enable_separate_option_page = false;

	/**
	 * Name of the Link that leads to the plugin's options page
	 *
	 * @var string
	 */
	protected $options_link_title = '';

	/**
	 * Title (HTML head) of the plugin's options page (if any)
	 *
	 * @var string
	 */
	protected $options_page_title = '';

	/**
	 * Plugin information and URLs for displaying in the options footer
	 *
	 * @var string
	 */
	protected $plugin_infos = array();

	/**
	 * Priority for the init action (init_plugin method)
	 *
	 * @var string
	 */
	protected $init_plugin_priority;

	/**
	 * Set of core plugin data (name, slug, version etc.)
	 *
	 * @var string[]
	 */
	protected $bootstrap_data = array();

	/**
	 * Admin notices to display
	 *
	 * @var mixed[]
	 */
	protected $admin_notices = array();

	/**
	 * Plugin slug
	 *
	 * @var string
	 */
	public $plugin_slug;

	/**
	 * Stable/Release version flag
	 *
	 * @var bool
	 */
	public $is_stable;

	/**
	 * Name of the custom field for storing plugin options
	 *
	 * @var string
	 */
	public $plugin_options_name;

	/**
	 * Gettext textdomain of plugin translations
	 *
	 * @var string
	 */
	public $textdomain;

	/**
	 * Translations loaded flag
	 *
	 * @var bool
	 */
	public $translations_loaded = false;

	/**
	 * Plugin directory (full path)
	 *
	 * @var string
	 */
	public $plugin_dir;

	/**
	 * Plugin filesystem directory (full path), may differ from $this->plugin_dir)
	 *
	 * @var string
	 */
	public $plugin_fs_dir;

	/**
	 * Main plugin file (full path)
	 *
	 * @var string
	 */
	public $plugin_main_file;

	/**
	 * Main plugin file (path relative to WP plugin dir)
	 *
	 * @var string
	 */
	public $plugin_main_file_rel;

	/**
	 * Handle for enqueuing the main backend CSS file
	 *
	 * @var string
	 */
	public $backend_css_handle;

	/**
	 * Handle for enqueuing the main backend JS file
	 *
	 * @var string
	 */
	public $backend_js_handle;

	/**
	 * Handle for enqueuing the main frontend CSS file
	 *
	 * @var string
	 */
	public $frontend_base_css_handle;

	/**
	 * Handle for enqueuing the main frontend JS file
	 *
	 * @var string
	 */
	public $frontend_base_js_handle;

	/**
	 * Utility object
	 *
	 * @var Settings_Helper
	 */
	public $settings_helper;

	/**
	 * Utility object
	 *
	 * @var General_Utils
	 */
	public $general_utils;

	/**
	 * Utility object
	 *
	 * @var String_Utils
	 */
	public $string_utils;

	/**
	 * Utility object
	 *
	 * @var Geo_Utils
	 */
	public $geo_utils;

	/**
	 * Utility object
	 *
	 * @var Template_Utils
	 */
	public $template_utils;

	/**
	 * Set of all utility class instances mentioned above
	 *
	 * @var object[]
	 */
	public $core_utils;

	/**
	 * Additional utility objects (will be merged with core utils)
	 *
	 * @var object[]
	 */
	public $utils = array();

	/**
	 * Debug object
	 *
	 * @var Debug
	 */
	public $debug;

	/**
	 * WP_Filesystem object
	 *
	 * @var \WP_Filesystem_Base
	 */
	public $wp_filesystem;

	/**
	 * WP_Filesystem object (alias)
	 *
	 * @var \WP_Filesystem_Base
	 */
	public $fs;

	/**
	 * Has the plugin been activated network-wide?
	 *
	 * @var bool
	 */
	public $is_network_activated;

	/**
	 * Constructor: Set plugin slug and dependent variables.
	 *
	 * @since 0.1
	 *
	 * @param string      $plugin_slug Plugin slug.
	 * @param string|bool $textdomain Plugin text domain (optional).
	 * @param string|bool $plugin_fs_dir Plugin filesystem directory (optional).
	 *
	 * @throws \Exception Exception thrown if plugin slug is missing.
	 */
	public function __construct( $plugin_slug, $textdomain = false, $plugin_fs_dir = false ) {
		if ( $plugin_slug ) {
			$this->plugin_slug            = $plugin_slug;
			$this->plugin_dir             = WP_PLUGIN_DIR . '/' . $plugin_slug;
			$this->plugin_fs_dir          = $plugin_fs_dir ? $plugin_fs_dir : $this->plugin_dir;
			$this->plugin_main_file       = $this->plugin_dir . '/' . $this->plugin_slug . '.php';
			$this->plugin_main_file_rel   = $this->plugin_slug . '/' . $this->plugin_slug . '.php';
			$this->plugin_options_name    = $plugin_slug . '_options';
			$this->default_plugin_options = $this->plugin_options;

			$this->plugin_infos = array(
				'plugin_main_file' => $this->plugin_main_file,
				'settings_page'    => '',
			);
		} else {
			throw new \Exception( 'inveris WP Free Plugin Core: Plugin slug (= directory name) not provided.' );
		}

		$this->is_stable  = preg_match( '/^[0-9]+\.[0-9]+(\.[0-9]+)?$/', static::PLUGIN_VERSION ) ? true : false;
		$this->textdomain = $textdomain;

		require_once ABSPATH . 'wp-admin/includes/plugin.php';
		require_once ABSPATH . 'wp-admin/includes/file.php';

		global $wp_filesystem;
		WP_Filesystem();
		$this->wp_filesystem = $wp_filesystem;
		$this->fs            = $wp_filesystem;

		if ( ! function_exists( 'is_plugin_active_for_network' ) ) {
			require_once ABSPATH . 'wp-admin/includes/plugin.php';
		}
		$this->is_network_activated = is_plugin_active_for_network( $this->plugin_main_file_rel );

		register_activation_hook( $this->plugin_main_file, array( $this, 'activate_plugin' ) );
		register_deactivation_hook( $this->plugin_main_file, array( $this, 'deactivate_plugin' ) );

		$this->bootstrap_data = array_merge(
			$this->bootstrap_data,
			array(
				'plugin_name'         => static::PLUGIN_NAME,
				'plugin_slug'         => $plugin_slug,
				'plugin_version'      => static::PLUGIN_VERSION,
				'plugin_prefix'       => static::PLUGIN_PREFIX,
				'public_prefix'       => static::PUBLIC_PREFIX,
				'plugin_dir'          => $this->plugin_dir,
				'plugin_fs_dir'       => $this->plugin_fs_dir,
				'plugin_main_file'    => $this->plugin_main_file,
				'plugin_options_name' => $this->plugin_options_name,
				'has_free_license'    => ! defined( 'static::FREE_LICENSE' ) || static::FREE_LICENSE,
			)
		);

		add_action( static::PLUGIN_PREFIX . 'update_plugin_options', array( $this, 'update_plugin_options' ), 10, 2 );
		add_action( static::PLUGIN_PREFIX . 'add_deferred_admin_notice', array( $this, 'add_deferred_admin_notice' ), 10, 4 );
		add_action( static::PLUGIN_PREFIX . 'dismiss_deferred_admin_notice', array( $this, 'dismiss_admin_notice' ), 10 );
		add_action( static::PLUGIN_PREFIX . 'admin_mail', array( $this, 'send_admin_mail' ), 10, 6 );

		add_action( 'wp_ajax_dismiss_admin_notice', array( $this, 'dismiss_admin_notice' ) );
	} // __construct

	/**
	 * Get values/objects from plugin options, bootstrap data, plugin infos
	 * or utils.
	 *
	 * @since 0.9
	 *
	 * @param string $key Option/Object name.
	 *
	 * @return mixed Requested Value/Object or false if nonexistent.
	 */
	public function __get( $key ) {
		$value = null;

		switch ( $key ) {
			case 'bootstrap_data':
				$value = $this->bootstrap_data;
				break;
			case 'plugin_infos':
				$value = $this->plugin_infos;
				break;
			case 'cpt_hooks':
				$value = $this->cpt_hooks;
				break;
			case 'utils':
				$value = $this->utils;
				break;
			default:
				if ( isset( $this->plugin_options[ $key ] ) ) {
					$value = $this->plugin_options[ $key ];
				}
				if ( isset( $this->bootstrap_data[ $key ] ) ) {
					$value = $this->bootstrap_data[ $key ];
				}
				if ( isset( $this->plugin_infos[ $key ] ) ) {
					$value = $this->plugin_infos[ $key ];
				}
				if ( isset( $this->utils[ $key ] ) ) {
					$value = $this->utils[ $key ];
				}
		}

		return apply_filters(
			'immonex_core_magic_get_value',
			is_null( $value ) ? false : $value,
			$key,
			! is_null( $value ),
			$this->plugin_slug
		);
	} // __get

	/**
	 * Return the current plugin debug level status (compatibility with legacy
	 * plugin core).
	 *
	 * @since 1.5.3
	 *
	 * @return int Debug level.
	 */
	public function is_debug() {
		return is_object( $this->debug ) ? $this->debug->get_debug_level() : 0;
	} // is_debug

	/**
	 * Perform activation tasks.
	 *
	 * @since 0.9
	 *
	 * @param bool $network_wide Indicate network wide activation.
	 */
	public function activate_plugin( $network_wide = false ) {
		// Pre-loading of translations is required on activation.
		$this->load_translations();

		if ( $network_wide ) {
			global $wpdb;

			// Retrieve all site IDs from this network (SQL query for compatibility reasons).
			// @codingStandardsIgnoreLine
			$site_ids = $wpdb->get_col( "SELECT blog_id FROM {$wpdb->blogs} WHERE site_id = {$wpdb->siteid}" );

			foreach ( $site_ids as $site_id ) {
				switch_to_blog( $site_id );
				$this->activate_plugin_single_site();
				restore_current_blog();
			}
		} else {
			$this->activate_plugin_single_site();
		}
	} // activate_plugin

	/**
	 * Perform activation tasks for a single site.
	 *
	 * @since 1.1.0
	 */
	protected function activate_plugin_single_site() {
		// Fetch plugin options and update version.
		$this->plugin_options = $this->fetch_plugin_options();

		if ( static::PLUGIN_VERSION !== $this->plugin_options['plugin_version'] ) {
			$this->plugin_options['plugin_version'] = static::PLUGIN_VERSION;
			update_option( $this->plugin_options_name, $this->plugin_options );
		}

		// Schedule frequent tasks.
		if ( ! wp_get_schedule( static::PLUGIN_PREFIX . 'do_daily' ) ) {
			wp_schedule_event( time(), 'daily', static::PLUGIN_PREFIX . 'do_daily' );
		}
		if ( ! wp_get_schedule( static::PLUGIN_PREFIX . 'do_weekly' ) ) {
			wp_schedule_event( time(), 'weekly', static::PLUGIN_PREFIX . 'do_weekly' );
		}
	} // activate_plugin_single_site

	/**
	 * Perform deactivation tasks.
	 *
	 * @since 0.9
	 */
	public function deactivate_plugin() {
		wp_clear_scheduled_hook( static::PLUGIN_PREFIX . 'do_daily' );
		wp_clear_scheduled_hook( static::PLUGIN_PREFIX . 'do_weekly' );
	} // deactivate_plugin

	/**
	 * Register plugin settings.
	 *
	 * @since 0.9
	 */
	public function register_plugin_settings() {
		// All plugin options are stored in one serialized array.
		register_setting(
			$this->plugin_options_name,
			$this->plugin_options_name,
			array(
				'sanitize_callback' => array( $this, 'sanitize_plugin_options' ),
			)
		);

		if ( $this->enable_separate_option_page ) {
			if ( 'settings' === static::OPTIONS_LINK_MENU_LOCATION ) {
				// Add options page link in WP default settings menu.
				add_options_page(
					$this->options_page_title,
					$this->options_link_title,
					$this->plugin_options_access_capability,
					$this->plugin_slug . '_settings',
					array( $this->settings_helper, 'render_page' )
				);
			} else {
				// Add options page link as submenu item.
				$options_menu_item = array(
					static::OPTIONS_LINK_MENU_LOCATION,
					$this->options_page_title,
					$this->options_link_title,
					$this->plugin_options_access_capability,
					$this->plugin_slug . '_settings',
					array( $this->settings_helper, 'render_page' ),
					900,
				);

				call_user_func_array( 'add_submenu_page', $options_menu_item );
			}
		}
	} // register_plugin_settings

	/**
	 * Start plugin initialization.
	 *
	 * @param int $init_plugin_priority Priority for the init action
	 *                                  (init_plugin method).
	 *
	 * @since 0.1
	 */
	public function init( $init_plugin_priority = 10 ) {
		$this->init_plugin_priority = $init_plugin_priority;
		add_action( 'plugins_loaded', array( $this, 'init_base' ) );
	} // init

	/**
	 * Perform core initialization tasks.
	 *
	 * @since 0.1
	 */
	public function init_base() {
		$this->load_translations();

		add_action( 'init', array( $this, 'init_plugin' ), $this->init_plugin_priority );
		add_action( 'widgets_init', array( $this, 'init_plugin_widgets' ) );
		add_action( 'admin_init', array( $this, 'init_plugin_admin' ) );
		add_action( 'admin_menu', array( $this, 'register_plugin_settings' ) );

		// Exclude plugin JS/CSS from Autoptimize "optimizations".
		add_filter( 'option_autoptimize_js_exclude', array( $this, 'autoptimize_exclude' ), 10, 2 );
		add_filter( 'option_autoptimize_css_exclude', array( $this, 'autoptimize_exclude' ), 10, 2 );

		// Add a filter for modifying the required user/role capability for
		// accessing and updating plugin options.
		add_filter( 'option_page_capability_' . $this->plugin_options_name, array( $this, 'get_plugin_options_access_capability' ) );

		$enable_separate_option_page = false;

		if (
			defined( 'static::OPTIONS_LINK_MENU_LOCATION' ) &&
			static::OPTIONS_LINK_MENU_LOCATION
		) {
			if ( 'settings' === static::OPTIONS_LINK_MENU_LOCATION ) {
				$this->plugin_infos['settings_page'] = wp_sprintf(
					'options-general.php?page=%s_settings',
					$this->plugin_slug
				);

				if ( empty( $this->options_link_title ) ) {
					$this->options_link_title = static::PLUGIN_NAME;
				}
			} else {
				$this->plugin_infos['settings_page'] = wp_sprintf(
					'admin.php?page=%s_settings',
					$this->plugin_slug
				);

				if ( empty( $this->options_link_title ) ) {
					$this->options_link_title = __( 'Settings', 'immonex-wp-free-plugin-core' );
				}
			}

			if ( empty( $this->options_page_title ) ) {
				$this->options_page_title = static::PLUGIN_NAME . ' - ' .
					__( 'Settings', 'immonex-wp-free-plugin-core' );
			}

			$enable_separate_option_page = true;
		}

		$this->enable_separate_option_page = apply_filters(
			// @codingStandardsIgnoreLine
			$this->plugin_slug . '_enable_separate_option_page',
			$enable_separate_option_page
		);

		if ( ! isset( $this->default_plugin_options['debug_level'] ) ) {
			$this->default_plugin_options['debug_level'] = 0;
		}

		if ( ! isset( $this->default_plugin_options['deferred_admin_notices'] ) ) {
			$this->default_plugin_options['deferred_admin_notices'] = array();
		}
	} // init_base

	/**
	 * Initialize the plugin (common).
	 *
	 * @since 0.1
	 *
	 * @param bool $fire_before_hook Flag to indicate if an action hook should fire
	 *                               before the actual method execution (optional,
	 *                               true by default).
	 * @param bool $fire_after_hook  Flag to indicate if an action hook should fire
	 *                               after the actual method execution (optional,
	 *                               true by default).
	 */
	public function init_plugin( $fire_before_hook = true, $fire_after_hook = true ) {
		if ( $fire_before_hook ) {
			do_action( 'immonex_core_before_init', $this->plugin_slug );
		}

		// Retrieve the plugin options (merge with default values).
		$this->plugin_options = $this->fetch_plugin_options();

		$this->plugin_options_access_capability = apply_filters(
			// @codingStandardsIgnoreLine
			$this->plugin_slug . '_plugin_options_access_capability',
			$this->plugin_options_access_capability
		);

		/**
		 * Check/Update the debug level.
		 */
		$this->debug          = new Debug( $this->plugin_options, $this->plugin_options_access_capability, $this->plugin_options_name, $this->plugin_slug );
		$this->plugin_options = $this->debug->maybe_update_debug_level( $this->plugin_options );
		$this->extend_plugin_infos();

		/**
		 * Instantiate helper and utility classes.
		 */
		$this->general_utils   = new General_Utils();
		$this->settings_helper = new Settings_Helper(
			$this->plugin_dir,
			$this->plugin_slug,
			$this->plugin_options_name,
			$this->plugin_infos,
			$this->plugin_options_access_capability
		);
		$this->string_utils    = new String_Utils();
		$this->geo_utils       = new Geo_Utils();
		$this->template_utils  = new Template_Utils(
			$this,
			! empty( $this->plugin_options['skin'] ) ?
				$this->plugin_options['skin'] :
				''
		);
		$this->color_utils     = new Color_Utils( $this );
		$this->mail_utils      = new Mail_Utils( $this->plugin_slug, $this->string_utils, $this->template_utils );

		$this->core_utils = array(
			'general'  => $this->general_utils,
			'settings' => $this->settings_helper,
			'string'   => $this->string_utils,
			'geo'      => $this->geo_utils,
			'mail'     => $this->mail_utils,
			'template' => $this->template_utils,
			'color'    => $this->color_utils,
		);

		if ( is_array( $this->utils ) && count( $this->utils ) > 0 ) {
			$this->utils = array_merge(
				$this->core_utils,
				$this->utils
			);
		} else {
			$this->utils = $this->core_utils;
		}

		if ( ! empty( $this->plugin_options['skin'] ) ) {
			// Check if skin folder still exists.
			$skins = $this->utils['template']->get_frontend_skins();
			if ( ! isset( $skins[ $this->plugin_options['skin'] ] ) ) {
				$this->plugin_options['skin'] = 'default';
				update_option( $this->plugin_options_name, $this->plugin_options );
			}
		}

		// Add WP-Cron-based actions.
		add_action( static::PLUGIN_PREFIX . 'do_daily', array( $this, 'do_daily' ) );
		add_action( static::PLUGIN_PREFIX . 'do_weekly', array( $this, 'do_weekly' ) );

		// Enqueue frontend CSS and JS files.
		add_action( 'wp_enqueue_scripts', array( $this, 'frontend_scripts_and_styles' ) );

		if ( $fire_after_hook ) {
			do_action( 'immonex_core_after_init', $this->plugin_slug );
		}
	} // init_plugin

	/**
	 * Initialize the plugin (admin/backend only).
	 *
	 * @since 0.1
	 *
	 * @param bool $fire_before_hook Flag to indicate if an action hook should fire
	 *                               before the actual method execution (optional,
	 *                               true by default).
	 * @param bool $fire_after_hook  Flag to indicate if an action hook should fire
	 *                               after the actual method execution (optional,
	 *                               true by default).
	 */
	public function init_plugin_admin( $fire_before_hook = true, $fire_after_hook = true ) {
		if ( $fire_before_hook ) {
			do_action( 'immonex_core_before_init_admin', $this->plugin_slug );
		}

		// @codingStandardsIgnoreLine
		$script = isset( $_SERVER['SCRIPT_NAME'] ) ? sanitize_file_name( basename( wp_unslash( $_SERVER['SCRIPT_NAME'] ), '.php' ) ) : '';

		if (
			( isset( $_GET['settings-updated'] ) && 'true' === $_GET['settings-updated'] ) &&
			( isset( $_GET['page'] ) && $this->plugin_slug . '_settings' === $_GET['page'] ) &&
			'options-general' !== $script
		) {
			$this->add_admin_notice(
				wp_sprintf(
					'<strong>%s</strong>',
					__( 'Settings saved.', 'immonex-wp-free-plugin-core' )
				),
				'success',
				'settings_updated'
			);
		}

		add_action( 'admin_enqueue_scripts', array( $this, 'admin_scripts_and_styles' ) );
		add_action( 'network_admin_notices', array( $this, 'display_network_admin_notices' ) );
		add_action( 'admin_notices', array( $this, 'display_admin_notices' ) );

		// Add a "Settings" link on the plugins page.
		if ( $this->enable_separate_option_page ) {
			add_filter(
				'plugin_action_links_' . $this->plugin_slug . '/' . $this->plugin_slug . '.php',
				array( $this->settings_helper, 'plugin_settings_link' )
			);
		}

		if (
			is_array( $this->plugin_options['deferred_admin_notices'] ) &&
			count( $this->plugin_options['deferred_admin_notices'] ) > 0
		) {
			// Show deferred admin notices until dismissed.
			foreach ( $this->plugin_options['deferred_admin_notices'] as $id => $admin_notice ) {
				$this->add_admin_notice( $admin_notice['message'], $admin_notice['type'], $id );
			}
		}

		if (
			! $this->plugin_options['plugin_version'] ||
			version_compare( $this->plugin_options['plugin_version'], static::PLUGIN_VERSION, '<' )
		) {
			// Plugin has been updated: redo activation.
			$this->activate_plugin();
		}

		if ( $fire_after_hook ) {
			do_action( 'immonex_core_after_init_admin', $this->plugin_slug );
		}
	} // init_plugin_admin

	/**
	 * Load and register widgets (to be overwritten in derived classes).
	 *
	 * @since 0.1
	 */
	public function init_plugin_widgets() {}

	/**
	 * Enqueue and localize frontend scripts and styles.
	 *
	 * @since 0.9
	 */
	public function frontend_scripts_and_styles() {
		/**
		 * Plugin base CSS
		 */
		$base_css_folder = '';
		foreach ( array( 'assets/css', 'assets', 'css' ) as $folder ) {
			if ( file_exists( trailingslashit( $this->plugin_dir ) . "{$folder}/frontend.css" ) ) {
				$base_css_folder = $folder;
				break;
			}
		}

		if ( $base_css_folder ) {
			$this->frontend_base_css_handle = static::PUBLIC_PREFIX . 'frontend';

			wp_enqueue_style(
				$this->frontend_base_css_handle,
				plugins_url( $this->plugin_slug . "/{$base_css_folder}/frontend.css" ),
				array(),
				$this->plugin_version
			);
		}

		/**
		 * Plugin base JS
		 */
		$base_js_folder = '';
		foreach ( array( 'assets/js', 'assets', 'js' ) as $folder ) {
			if ( file_exists( trailingslashit( $this->plugin_dir ) . "{$folder}/frontend.js" ) ) {
				$base_js_folder = $folder;
				break;
			}
		}

		if ( $base_js_folder ) {
			$this->frontend_base_js_handle = static::PUBLIC_PREFIX . 'frontend';

			wp_register_script(
				$this->frontend_base_js_handle,
				plugins_url( $this->plugin_slug . "/{$base_js_folder}/frontend.js" ),
				array( 'jquery' ),
				$this->plugin_version,
				true
			);
			wp_enqueue_script( $this->frontend_base_js_handle );
		}

		if ( ! empty( $this->plugin_options['skin'] ) ) {
			$lookup_basenames = array( 'index', 'extend' );

			foreach ( $lookup_basenames as $basename ) {
				/**
				 * Skin CSS
				 */
				$skin_css = $this->utils['template']->locate_template_file( "css/{$basename}.css" );
				if ( $skin_css ) {
					$skin_css_url = $this->utils['template']->get_template_file_url( $skin_css );

					if ( $skin_css_url ) {
						$skin_css_deps = array();
						if ( wp_style_is( $this->frontend_base_css_handle ) ) {
							$skin_css_deps[] = $this->frontend_base_css_handle;
						}

						$handle = static::PUBLIC_PREFIX . 'skin';
						if ( 'index' !== $basename ) {
							$handle .= "-{$basename}";
						}

						wp_enqueue_style(
							$handle,
							$skin_css_url,
							$skin_css_deps,
							$this->plugin_version
						);
					}
				}

				/**
				 * Skin JS
				 */
				$skin_js = $this->utils['template']->locate_template_file( "js/{$basename}.js" );
				if ( $skin_js ) {
					$skin_js_url = $this->utils['template']->get_template_file_url( $skin_js );

					if ( $skin_js_url ) {
						$skin_js_deps = array( 'jquery' );
						if ( wp_script_is( $this->frontend_base_js_handle ) ) {
							$skin_js_deps[] = $this->frontend_base_js_handle;
						}

						$handle = static::PUBLIC_PREFIX . 'skin';
						if ( 'index' !== $basename ) {
							$handle .= "-{$basename}";
						}

						wp_register_script(
							$handle,
							$skin_js_url,
							$skin_js_deps,
							$this->plugin_version,
							true
						);
						wp_enqueue_script( $handle );
					}
				}
			}
		}
	} // frontend_scripts_and_styles

	/**
	 * Retrieve plugin options and filter out invalid/outdated data.
	 *
	 * @since 0.1
	 *
	 * @param mixed[]|bool $defaults Default/Valid plugin options (optional).
	 *
	 * @return mixed[] Array of current plugin options.
	 */
	public function fetch_plugin_options( $defaults = false ) {
		/**
		 * Cache flush normally not required - reactivate if problems should
		 * occur in the future.
		 * wp_cache_flush();
		 */
		$plugin_options = get_option( $this->plugin_options_name );
		$new_options    = $defaults ? $defaults : $this->default_plugin_options;

		if ( is_array( $plugin_options ) ) {
			foreach ( $plugin_options as $key => $value ) {
				if ( isset( $new_options[ $key ] ) ) {
					$new_options[ $key ] = $value;
				}
			}
		}

		return $new_options;
	} // fetch_plugin_options

	/**
	 * Update changed plugin options (action callback).
	 *
	 * @since 1.2.2
	 *
	 * @param mixed[] $update_options Subset of options to be updated.
	 * @param string  $context        Update context (optional).
	 */
	public function update_plugin_options( $update_options, $context = '' ) {
		// (Re)fetch current plugin options.
		$this->plugin_options = $this->fetch_plugin_options();
		$options_changed      = false;

		foreach ( $this->plugin_options as $key => $old_value ) {
			if (
				isset( $update_options[ $key ] )
				&& $update_options[ $key ] !== $old_value
			) {
				$this->plugin_options[ $key ] = $update_options[ $key ];
				$options_changed              = true;
			}
		}

		if ( $options_changed ) {
			update_option( $this->plugin_options_name, $this->plugin_options );
		}
	} // update_plugin_options

	/**
	 * Possibly set a new capability for accessing and updating plugin options.
	 *
	 * @since 0.9
	 *
	 * @param string $cap Current capability.
	 *
	 * @return string (Possibly) updated capability.
	 */
	public function get_plugin_options_access_capability( $cap ) {
		return $this->plugin_options_access_capability;
	} // get_plugin_options_access_capability

	/**
	 * Enqueue and localize backend JavaScript and CSS code (callback).
	 *
	 * @since 0.1
	 *
	 * @param string $hook_suffix The current admin page.
	 */
	public function admin_scripts_and_styles( $hook_suffix ) {
		$ns_split            = explode( '\\', __NAMESPACE__ );
		$core_version        = array_pop( $ns_split );
		$core_version_handle = str_replace( '_', '-', $core_version );
		if ( 'V' === $core_version_handle[0] ) {
			$core_version_handle = substr( $core_version, 1 );
			$core_version_semver = str_replace( '-', '.', $core_version_handle );
		} else {
			$dev_number          = str_replace( 'DEV-', '', $core_version_handle );
			$core_version_semver = "{$dev_number}.0.0-alpha";
		}
		$core_handle = "{$this->plugin_slug}-backend-core-{$core_version_handle}";

		/**
		 * Load core backend CSS.
		 */
		wp_enqueue_style(
			$core_handle,
			plugins_url( $this->plugin_slug . "/vendor/immonex/wp-free-plugin-core/src/{$core_version}/css/backend.css" ),
			array(),
			$this->core_version_semver
		);

		/**
		 * Load plugin-specific CSS if existent.
		 */
		$base_css_folder = '';
		foreach ( array( 'assets/css', 'assets', 'css' ) as $folder ) {
			if ( file_exists( trailingslashit( $this->plugin_dir ) . "{$folder}/backend.css" ) ) {
				$base_css_folder = $folder;
				break;
			}
		}

		if ( $base_css_folder ) {
			$this->backend_css_handle = static::PUBLIC_PREFIX . 'backend';

			wp_enqueue_style(
				$this->backend_css_handle,
				plugins_url( $this->plugin_slug . "/{$base_css_folder}/backend.css" ),
				array(),
				$this->plugin_version
			);
		}

		/**
		 * Load core backend JS first.
		 */
		wp_register_script(
			$core_handle,
			plugins_url( $this->plugin_slug . "/vendor/immonex/wp-free-plugin-core/src/{$core_version}/js/backend.js" ),
			array( 'jquery' ),
			$core_version_semver,
			true
		);
		wp_enqueue_script( $core_handle );

		$media_fields = $this->settings_helper->get_media_fields();

		wp_localize_script(
			$core_handle,
			'iwpfpc_params',
			array(
				'core_version'                    => $core_version_semver,
				'plugin_slug'                     => $this->plugin_slug,
				'ajax_url'                        => get_admin_url() . 'admin-ajax.php',
				'media_fields'                    => $media_fields,
				'default_media_frame_title'       => __( 'Image Selection', 'immonex-wp-free-plugin-core' ),
				'default_media_frame_button_text' => __( 'Apply selection', 'immonex-wp-free-plugin-core' ),
			)
		);

		/**
		 * Load plugin-specific backend JS if existent.
		 */
		$base_js_folder = '';
		foreach ( array( 'assets/js', 'assets', 'js' ) as $folder ) {
			if ( file_exists( trailingslashit( $this->plugin_dir ) . "{$folder}/backend.js" ) ) {
				$base_js_folder = $folder;
				break;
			}
		}

		if ( $base_js_folder ) {
			$this->backend_js_handle = static::PUBLIC_PREFIX . 'backend';

			wp_register_script(
				$this->backend_js_handle,
				plugins_url( $this->plugin_slug . "/{$base_js_folder}/backend.js" ),
				array( 'jquery' ),
				$this->plugin_version,
				true
			);
			wp_enqueue_script( $this->backend_js_handle );
		}

		if ( ! empty( $media_fields ) ) {
			wp_enqueue_media();
		}
	} // admin_scripts_and_styles

	/**
	 * Display current messages for the network admin (callback).
	 *
	 * @since 1.5.2
	 */
	public function display_network_admin_notices() {
		$this->display_admin_notices( 'network' );
	} // display_network_admin_notices

	/**
	 * Display current administrative messages (callback).
	 *
	 * @param string $target "network" for network admin only notices, empty by default.
	 *
	 * @since 0.1
	 */
	public function display_admin_notices( $target = '' ) {
		$admin_notices = array_filter(
			$this->admin_notices,
			function ( $notice ) use ( $target ) {
				return $notice['target'] === $target;
			}
		);

		if ( count( $admin_notices ) > 0 ) {
			// Display admin messages.
			$dismissables_cnt = 0;
			foreach ( $admin_notices as $id => $notice ) {
				if ( $notice['is_dismissable'] ) {
					$dismissables_cnt++;
				}
				if ( $dismissables_cnt > 1 ) {
					break;
				}
			}

			foreach ( $admin_notices as $id => $notice ) {
				$message = $notice['message'];
				$classes = array(
					'notice',
					'notice-' . $notice['type'],
					$this->plugin_slug . '-notice',
				);

				if ( $notice['is_dismissable'] ) {
					$classes[] = 'is-dismissible';
				}

				echo wp_sprintf(
					'<div class="%s" data-notice-id="%s"><p>%s</p></div>' . PHP_EOL,
					implode( ' ', $classes ),
					$id,
					$message
				);

				// Remove displayed message.
				unset( $this->admin_notices[ $id ] );
			}
		}
	} // display_admin_notices

	/**
	 * Sanitize plugin options before saving.
	 *
	 * @since 0.9
	 *
	 * @param array $input Submitted form data.
	 *
	 * @return array Valid inputs.
	 */
	public function sanitize_plugin_options( $input ) {
		$valid       = array();
		$current_tab = $this->settings_helper->get_current_tab();
		$tab_fields  = $this->settings_helper->get_tab_fields( $current_tab );

		if ( count( $tab_fields ) > 0 ) {
			foreach ( $tab_fields as $name => $field ) {
				$exists = isset( $input[ $name ] );

				$value = '';
				if ( $exists ) {
					$value = $input[ $name ];
				} elseif ( isset( $field['default'] ) ) {
					$value = $field['default'];
				} elseif ( 'checkbox' !== $field['type'] ) {
					continue;
				}

				if ( isset( $field['force_type'] ) ) {
					settype( $value, $field['force_type'] );
				}

				if (
					! empty( $field['max_length'] ) &&
					is_string( $value ) &&
					strlen( $value ) > $field['max_length']
				) {
					$value = substr( trim( $value ), 0, $field['max_length'] );
				}

				if (
					isset( $field['min'] ) &&
					(float) $value < $field['min']
				) {
					$value = $field['min'];
				}

				if (
					isset( $field['max'] ) &&
					(float) $value > $field['max']
				) {
					$value = $field['max'];
				}

				switch ( $field['type'] ) {
					case 'select':
						if (
							$exists &&
							isset( $field['options'][ $value ] )
						) {
							$valid[ $name ] = $value;
						} elseif ( isset( $field['options'][0] ) ) {
							$valid[ $name ] = array_keys( $field['options'] )[0];
						}
						break;
					case 'checkbox':
						$valid[ $name ] = $exists;
						break;
					case 'checkbox_group':
						if ( $exists ) {
							$valid[ $name ] = $value;
						} else {
							$valid[ $name ] = array();
						}
						break;
					case 'wysiwyg':
						$valid[ $name ] = wp_kses_post( trim( $value ) );
						break;
					case 'textarea':
						$valid[ $name ] = sanitize_textarea_field( $value );
						break;
					default:
						// Normal text fields.
						$valid[ $name ] = sanitize_text_field( $value );
				}
			}
		}

		$options = apply_filters(
			// @codingStandardsIgnoreLine
			$this->plugin_slug . '_options_before_save',
			$this->settings_helper->merge_options( $this->plugin_options, $valid ),
			$this->plugin_options
		);

		return $options;
	} // sanitize_plugin_options

	/**
	 * Extend plugin information for displaying on the options page/tab.
	 *
	 * @since 0.9
	 */
	public function extend_plugin_infos() {
		$this->plugin_infos = array_merge(
			$this->plugin_infos,
			array(
				'name'             => defined( 'static::PLUGIN_NAME' ) ? static::PLUGIN_NAME : '',
				'prefix'           => defined( 'static::PLUGIN_PREFIX' ) ? static::PLUGIN_PREFIX : '',
				'debug_level'      => $this->is_debug(),
				'logo_link_url'    => defined( 'static::PLUGIN_HOME_URL' ) ? static::PLUGIN_HOME_URL : '',
				'footer'           => array(),
				'has_free_license' => ! defined( 'static::FREE_LICENSE' ) || static::FREE_LICENSE,
			)
		);

		$this->plugin_infos['footer'] = $this->get_plugin_footer_infos();
	} // extend_plugin_infos

	/**
	 * Delete a dismissible admin notice (callback).
	 *
	 * @param string $notice_id Notice ID (direct do_action calls only).
	 *
	 * @since 1.0.0
	 */
	public function dismiss_admin_notice( $notice_id = '' ) {
		if ( $notice_id ) {
			// Call with plugin slug in action name.
			$plugin_slug = $this->plugin_slug;
			$is_ajax     = false;
		} else {
			// AJAX request: Get notice ID and plugin slug from POST variables.
			// @codingStandardsIgnoreStart
			$notice_id   = isset( $_POST['notice_id'] ) ? sanitize_key( $_POST['notice_id'] ) : false;
			$plugin_slug = isset( $_POST['plugin_slug'] ) ? sanitize_key( $_POST['plugin_slug'] ) : false;
			// @codingStandardsIgnoreEnd

			$is_ajax = true;
		}
		if ( ! $notice_id || ! $plugin_slug ) {
			return;
		}

		if (
			$plugin_slug === $this->plugin_slug &&
			isset( $this->plugin_options['deferred_admin_notices'][ $notice_id ] )
		) {
			unset( $this->plugin_options['deferred_admin_notices'][ $notice_id ] );
			update_option( $this->plugin_options_name, $this->plugin_options );

			if ( $is_ajax ) {
				wp_die();
			}
		}
	} // dismiss_admin_notice

	/**
	 * Compile arbitrary plugin information and doc/support links if given.
	 *
	 * @since 0.9
	 *
	 * @return string[] Array of info/link elements.
	 */
	protected function get_plugin_footer_infos() {
		$infos = array();

		if ( defined( 'static::PLUGIN_VERSION' ) ) {
			$infos[] = 'Version <strong>' . static::PLUGIN_VERSION . '</strong>';
		}

		if (
			defined( 'static::PLUGIN_DOC_URLS' ) &&
			count( static::PLUGIN_DOC_URLS ) > 0
		) {
			$infos[] = $this->get_language_link(
				static::PLUGIN_DOC_URLS,
				__( 'Documentation', 'immonex-wp-free-plugin-core' )
			);
		}

		if (
			defined( 'static::PLUGIN_SUPPORT_URLS' ) &&
			count( static::PLUGIN_SUPPORT_URLS ) > 0
		) {
			$infos[] = $this->get_language_link(
				static::PLUGIN_SUPPORT_URLS,
				'Support'
			);
		}

		if (
			defined( 'static::PLUGIN_DEV_URLS' ) &&
			count( static::PLUGIN_DEV_URLS ) > 0
		) {
			$infos[] = $this->get_language_link(
				static::PLUGIN_DEV_URLS,
				__( 'Development', 'immonex-wp-free-plugin-core' )
			);
		}

		return $infos;
	} // get_plugin_footer_infos

	/**
	 * Add an administrative message.
	 *
	 * @since 0.1
	 *
	 * @param string $message Message to display.
	 * @param string $type    Message type: "success" (default), "info", "warning", "error" or
	 *                        "network info/warning/error" for network admin notices.
	 * @param string $id      Message ID (required for deferred messages only).
	 */
	protected function add_admin_notice( $message, $type = 'success', $id = '' ) {
		$target = '';

		if ( false !== strpos( $type, ' ' ) ) {
			$type_split = explode( ' ', $type );
			$target     = 'network' === $type_split[0] ? 'network' : '';
			$type       = $type_split[0];
		}

		if ( ! in_array( $type, array( 'success', 'info', 'warning', 'error' ), true ) ) {
			$type = 'info';
		}

		$notice = array(
			'message'        => $message,
			'type'           => $type,
			'is_dismissable' => $id ? true : false,
			'target'         => $target,
		);

		$this->admin_notices[ $id ? $id : uniqid() ] = $notice;
	} // add_admin_notice

	/**
	 * Add a "deferred" administrative message that will be saved as part of the
	 * plugin options (also used as action callback).
	 *
	 * @since 0.3.6
	 *
	 * @param string $message Message to display.
	 * @param string $type    Message type: "success" (default), "info", "warning", "error" or
	 *                        "network info/warning/error" for network admin notices.
	 * @param string $context Message context (e.g. if called as action hook callback; optional).
	 * @param string $id      Optional message ID for deferred messages.
	 */
	public function add_deferred_admin_notice( $message, $type = 'success', $context = '', $id = '' ) {
		$raw_type = str_replace( 'network ', '', $type );

		if ( ! in_array( $raw_type, array( 'success', 'info', 'warning', 'error' ), true ) ) {
			$type = 'info';
		}

		// (Re)fetch current plugin options.
		$this->plugin_options = $this->fetch_plugin_options();

		if ( empty( $this->plugin_options['deferred_admin_notices'] ) ) {
			$this->plugin_options['deferred_admin_notices'] = array();
		} else {
			foreach ( $this->plugin_options['deferred_admin_notices'] as $notice ) {
				if ( $notice['message'] === $message ) {
					return;
				}
			}
		}

		$this->plugin_options['deferred_admin_notices'][ $id ? $id : uniqid() ] = array(
			'message' => $message,
			'type'    => $type,
		);

		update_option( $this->plugin_options_name, $this->plugin_options );
	} // add_deferred_admin_notice

	/**
	 * Send an admin info mail (callback).
	 *
	 * @since 1.5.0
	 *
	 * @param string               $subject       Subject.
	 * @param string|string[]      $body          Mail body (plain text only or HTML and text).
	 * @param string[]             $headers       Headers.
	 * @param string[]             $attachments   Attachment files (absolute paths).
	 * @param mixed[]              $template_data Data/Parameters for rendering the default HTML frame template.
	 * @param bool|string|string[] $to            Recipient(s) - optional, defaults to false (site/network admin).
	 */
	public function send_admin_mail( $subject, $body, $headers = array(), $attachments = array(), $template_data = array(), $to = false ) {
		if ( empty( $to ) ) {
			$to = get_option( 'admin_email' );
		}

		$template_data['preset']    = 'admin_info';
		$template_data['bootstrap'] = $this->bootstrap_data;

		$this->mail_utils->send( $to, $subject, $body, $headers, $attachments, $template_data );
	} // send_admin_mail

	/**
	 * Exclude plugin JS/CSS from Autoptimize "optimizations".
	 *
	 * @since 1.3.2
	 *
	 * @param string $value  Current exclusion patterns/terms.
	 * @param string $option Option name.
	 *
	 * @return string Extended list of exclusion patterns/terms.
	 */
	public function autoptimize_exclude( $value, $option ) {
		if ( false === strpos( $value, 'immonex' ) ) {
			return implode( ', ', array( $value, 'immonex' ) );
		}

		return $value;
	} // autoptimize_exclude

	/**
	 * Perform daily tasks.
	 *
	 * @since 1.6.0
	 */
	public function do_daily() {
		do_action( 'immonex_core_do_daily', $this->plugin_slug );
	} // do_daily

	/**
	 * Perform weekly tasks.
	 *
	 * @since 1.6.0
	 */
	public function do_weekly() {
		do_action( 'immonex_core_do_weekly', $this->plugin_slug );
	} // do_daily

	/**
	 * Show error messages during plugin activation.
	 *
	 * @since 0.2
	 *
	 * @param string $message Options form values.
	 * @param int    $errno Error number.
	 */
	protected function br_trigger_error( $message, $errno ) {
		if ( isset( $_GET['action'] ) && 'error_scrape' === $_GET['action'] ) {
			echo '<strong>' . $message . '</strong>';
			exit;
		} else {
			// @codingStandardsIgnoreLine
			trigger_error( $message, $errno );
		}
	} // br_trigger_error

	/**
	 * Load translations.
	 *
	 * @param bool $force_reload Maybe force a reload if translations have already
	 *                           been loaded (optional).
	 *
	 * @since 0.1
	 */
	protected function load_translations( $force_reload = false ) {
		if ( $this->translations_loaded && ! $force_reload ) {
			return;
		}

		$domain       = $this->textdomain ? $this->textdomain : $this->plugin_slug;
		$locale       = get_user_locale();
		$base_version = basename( __DIR__ );

		/**
		 * Load plugin base translations first.
		 */
		load_textdomain(
			'immonex-wp-free-plugin-core',
			wp_sprintf(
				'%s/%s/vendor/immonex/wp-free-plugin-core/src/%s/languages/immonex-wp-free-plugin-core-%s.mo',
				WP_PLUGIN_DIR,
				$this->plugin_slug,
				$base_version,
				$locale
			)
		);

		/**
		 * Load plugin translations.
		 */
		$always_load_global_translations = apply_filters(
			// @codingStandardsIgnoreLine
			$this->plugin_slug . '_always_load_global_translations',
			false
		);

		if ( $this->is_stable || $always_load_global_translations ) {
			load_plugin_textdomain( $domain, false, $this->plugin_slug . '/languages' );
		} else {
			$local_mo_file = wp_sprintf(
				'%s/%s/languages/%s-%s.mo',
				WP_PLUGIN_DIR,
				$this->plugin_slug,
				$this->textdomain,
				$locale
			);
			if ( file_exists( $local_mo_file ) ) {
				load_textdomain( $domain, $local_mo_file );
			}
		}

		$this->translations_loaded = true;
	} // load_translations

	/**
	 * Get a link tag related to the current language (or the default url).
	 *
	 * @since 0.9
	 *
	 * @param string[] $urls Array of URLs with language code as keys.
	 * @param string   $link_text Link text.
	 *
	 * @return string HTML link tag.
	 */
	private function get_language_link( $urls, $link_text ) {
		if ( empty( $urls ) ) {
			return '';
		}

		$lang = substr( get_user_locale(), 0, 2 );
		$href = empty( $urls[ $lang ] ) ? array_values( $urls )[0] : $urls[ $lang ];

		return wp_sprintf(
			'<a href="%1$s" target="_blank">%2$s</a>',
			$href,
			$link_text
		);
	} // get_language_link

} // Base
